from flask import Flask, request, g, session, redirect, abort
from datetime import datetime
from flask.ext.session import Session

app = Flask(__name__)
app.config.from_envvar('WEB_CONF')
Session(app)

@app.before_request
def before_request():
   if 'username' in session:
      username = session['username']
      g.user = app.config['AUTH_SERVICE'].getUser(username)
   authenticated = 'user' in g and g.user is not None

   if request.path.startswith('/assets'):
      return
   if (request.path == '/' or request.path == '/logout') and not authenticated:
      return redirect('/login')
   if request.path != '/login' and not authenticated:
      abort(401)



import io,json
import requests
import urllib
import uuid
from datetime import datetime
from pytz import timezone
from enum import Enum
from flask import render_template, redirect, request, url_for, escape, flash, g, abort, make_response, Response, stream_with_context
from .app import app

from duckpond.data.sparql import SPARQL

def curie(value):
   if value.find(':')<0:
      return "schema:"+value
   else:
      return value

def value(v,url=False):
   if type(v)==int:
      return str(v)
   if type(v)==float:
      return str(v)
   if type(v)==bool:
      return 'true' if v else 'false'
   if url:
      return '<' + v + '>'
   s = str(v)
   start = 0
   end = 0
   literal = io.StringIO()
   length = len(s)
   literal.write('"')
   for c in str(v):
      if c=='\\':
         literal.write('\\\\')
      elif c=='"':
         literal.write('\\"')
      elif c=='\u000a':
         literal.write('\\n')
      elif c=='\u000d':
         literal.write('\\r')
      else:
         literal.write(c)
   literal.write('"')
   return literal.getvalue()

def uri(s):
   return s[1:-1] if s[0]=='<' and s[-1]=='>' else s

def literal(s):
   return s[1:-1] if s[0]=='"' and s[-1]=='"' else s

def shorten(s):
   schema = 'http://schema.org/'
   return s[len(schema):] if s.find(schema)==0 else s

def expand(s):
   if s.find('schema:')==0:
      return 'http://schema.org/' + s[7:]
   if s.find(':')<0:
      return 'http://schema.org/' + s
   raise KeyError('Unknown prefix: '+s[0:s.find(':')])


class Service:

   class AuthMethod(Enum):
      basic = 1
      digest = 2

   def __init__(self,endpoints,cache=None):
      self.endpoints = {}
      self.endpoints['query'] = endpoints['query'] if 'query' in endpoints else None
      self.endpoints['update'] = endpoints['update'] if 'update' in endpoints else None
      self.cache = cache
      self.auth = self.authenticate('anonymous')

   def authenticate(self,user,password='',method=AuthMethod.basic):
      if (method is Service.AuthMethod.basic):
         self.auth = requests.auth.HTTPBasicAuth(user, password)
      elif (method is Service.AuthMethod.digest):
         self.auth = requests.auth.HTTPDigestAuth(user, password)
      else:
         raise Exception('Unsupported authentication method')

   def put(self,data,graph=None):
      uri = self.endpoints['update']
      if graph is not None:
         uri = uri + "?context=" + urllib.parse.quote('<'+graph+'>')

      if isinstance(data,str):
         data = data.encode('utf-8')

      req = requests.put(uri,data=data,headers={'content-type':'text/turtle; charset=utf-8'},auth=self.auth)

      #print(req.status_code)

      if (req.status_code<200 or req.status_code>=300):
         raise IOError('Cannot put data to uri <{}>, status={}'.format(uri,req.status_code))

   def post(self,data,graph=None):
      uri = self.endpoints['update']
      if graph is not None:
         uri = uri + "?context=" + urllib.parse.quote('<'+graph+'>')

      if isinstance(data,str):
         data = data.encode('utf-8')

      req = requests.post(uri,data=data,headers={'content-type':'text/turtle; charset=utf-8'},auth=self.auth)

      if (req.status_code<200 or req.status_code>=300):
         raise IOError('Cannot post data to uri <{}>, status={}'.format(uri,req.status_code))

   def exists(self,q,limit=None,graph=None):
      return self.query(q,limit=limit,graph=graph)

   def query(self,q,limit=None,graph=None):
      params = {'query':str(q)}
      if limit is not None:
         params['limit'] = limit
      if graph is not None:
         params['context'] = '<' + graph + '>'

      #print(str(q))

      req = requests.get(self.endpoints['query'],params=params,headers={'accept':'application/json'},auth=self.auth)

      if (req.status_code>=200 or req.status_code<300):
         data = json.loads(req.text)
         return data
      else:
         raise IOError('Cannot query to uri <{}>, status={}'.format(self.service,req.status_code))

   def update(self,q,limit=None,graph=None):
      params = {'query':str(q)}
      if limit is not None:
         params['limit'] = limit
      if graph is not None:
         params['context'] = '<' + graph + '>'

      #print(str(q))

      req = requests.post(self.endpoints['query'],params=params,headers={'accept':'application/json'},auth=self.auth)

      if (req.status_code>=200 or req.status_code<300):
         #print(req.text)
         data = json.loads(req.text)
         return data
      else:
         raise IOError('Cannot query to uri <{}>, status={}'.format(self.service,req.status_code))

   def retrieve(self,graph=None):
      uri = self.endpoints['update']
      params = {}
      if graph is not None:
         params['context'] = '<' + graph + '>'

      req = requests.get(uri,params=params,headers={'accept':'application/json'},auth=self.auth)

      if (req.status_code>=200 or req.status_code<300):
         data = json.loads(req.text)
         return data
      else:
         raise IOError('Cannot retrieve uri <{}>, status={}'.format(self.service,req.status_code))

   def deleteGraph(self,graph):
      uri = self.endpoints['update']
      params = {'context' : '<' + graph + '>' }

      req = requests.delete(uri,params=params,auth=self.auth)

      if (req.status_code>=200 or req.status_code<300):
         return True
      else:
         raise IOError('Cannot delete uri <{}>, status={}'.format(self.service,req.status_code))

   def deleteProperties(self,subject,property=None,object=None,graph=None):
      uri = self.endpoints['update']
      params = {'subj' : '<' + subject + '>' if subject[0]!='_' else subject }
      if property is not None:
         params['pred'] = '<' + expand(property) + '>'
      if object is not None:
         params['obj'] = object
      if graph is not None:
         params['context'] = '<' + graph + '>'

      req = requests.delete(uri,params=params,auth=self.auth)

      if (req.status_code>=200 or req.status_code<300):
         return True
      else:
         raise IOError('Cannot delete uri <{}>, status={}'.format(self.service,req.status_code))


class TurtleSerializer:
   def __init__(self):
      self.output = io.StringIO()
      self.urltypes = set(['url','contentUrl'])

   def __str__(self):
      return self.output.getvalue()

   def start(self):
      self.output.write('@prefix schema: <http://schema.org/> .\n')

   def end(self):
      return self.output.getvalue()

   def about(self,subject,data):
      if subject[0]=='_':
         self.output.write(subject)
         self.output.write('\n')
      else:
         self.output.write('<' + subject + '>\n')
      if '@type' in data:
         self.output.write('  a ' + curie(data['@type'] + ';\n'))
      first = True
      queue = []
      for name in data:
         if name == '@context' or name=='@type' or name=='@id':
            continue;
         if isinstance(data[name],dict):
            if not first:
               self.output.write(';\n')
            self.output.write('  ')
            self.output.write(curie(name))
            self.output.write(' ')
            if '@id' in data[name]:
               id = data[name]['@id']
               if id[0]=='_':
                  self.output.write(id)
               else:
                  self.output.write('<' + id + '>')
               queue.append(data[name])
            else:
               self.blankNode(data[name])
         elif isinstance(data[name],list):
            for obj in data[name]:
               if not first:
                  self.output.write(';\n')
               self.output.write('  ')
               self.output.write(curie(name))
               self.output.write(' ')
               if isinstance(obj,dict):
                  if '@id' in obj:
                     id = obj['@id']
                     if id[0]=='_':
                        self.output.write(id)
                     else:
                        self.output.write('<' + id + '>')
                     queue.append(obj)
                  else:
                     self.blankNode(obj)
               else:
                  self.output.write(value(obj,url=name in self.urltypes))
         else:
            if not first:
               self.output.write(';\n')
            self.output.write('  {0} {1}'.format(curie(name),value(data[name],url=name in self.urltypes)))
         first = False
      self.output.write('.\n')
      for o in queue:
         subject = o['@id']
         self.about(subject,o)
      #print(str(self))

   def blankNode(self,data):
      self.output.write('[')
      if '@type' in data:
         self.output.write(' a ' + curie(data['@type'] + ';\n'))
      first = True
      for name in data:
         if name == '@context' or name=='@type' or name=='@id':
            continue;
         if not first:
            self.output.write(';\n')
         if isinstance(data[name],dict):
            self.output.write('  ')
            self.output.write(curie(name))
            self.blankNode(data[name])
         else:
            self.output.write('  {0} {1}'.format(curie(name),value(data[name],url=name in self.urltypes)))
         first = False
      self.output.write(']')

def ask(data):
   turtle = io.StringIO()
   turtle.write('PREFIX schema: <http://schema.org/>\n')
   turtle.write('ASK { ?s \n')
   first = True
   for name in data:
      if name == '@context':
         continue;
      if not first:
         turtle.write(';\n')
      if name == '@type':
         turtle.write('  a '+curie(name))
      else:
         turtle.write('  {0} {1}'.format(curie(name),value(data[name],url=name=='url')))
      first = False
   turtle.write('\n}')
   s = turtle.getvalue()
   turtle.close()
   return s

def toJSONLD(subject,quads):
   objs = { subject : { '@context' : 'http://schema.org/', '@id' : subject}}

   toResolve = []
   for quad in quads:
      s = uri(quad[0])
      p = uri(quad[1])
      o = quad[2]
      obj = None
      if s in objs:
         obj = objs[s]
      else:
         obj = {'@id' : s}
         objs[s] = obj
      if p=='http://www.w3.org/1999/02/22-rdf-syntax-ns#type':
         obj['@type'] = shorten(uri(o))
      else:
         name = shorten(p)
         value = literal(uri(o))
         if name=='hasPart' or name=='associatedMedia' or name=='author' or value[0:2]=='_:':
            toResolve.append({ 'subject': obj, 'property': name, 'id': value})
         elif name in obj:
            if type(obj[name])!=list:
               l = [obj[name]]
               obj[name] = l
            obj[name].append(value)
         else:
            obj[name] = value

   for resolve in toResolve:
      name = resolve['property']
      target = resolve['subject']
      value = objs.get(resolve['id'])
      if value is None:
         continue
      if name=='hasPart' or name=='associatedMedia':
         if name not in target:
            target[name] = []
         target[name].append(value)
      elif name in target:
         if type(target[name])!=list:
            l = [target[name]]
            target[name] = l
         target[name].append(value)
      else:
         target[name] = value
   return objs[subject] if subject in objs else None

def now():
   return datetime.now(tz=timezone(app.config['TIMEZONE'] if 'TIMEZONE' in app.config else 'UTC')).isoformat()

def contentSubject(url):
   service = app.config['UPDATE_SERVICE']
   q = SPARQL() \
         .start({ 'schema' : 'http://schema.org/'}) \
         .select(['s']) \
         .where('?s schema:url <{}>'.format(url) )
   subjects = service.query(q)
   #print(subjects)
   return uri(subjects['values'][0][0]) if len(subjects['values'])>0 else None

def mediaSubject(s,url):
   service = app.config['UPDATE_SERVICE']
   q = SPARQL() \
         .start({ 'schema' : 'http://schema.org/'}) \
         .select(['s','p']) \
         .where('<{0}> ?p ?s . ?s schema:contentUrl <{1}>'.format(s,url) )
   subjects = service.query(q)
   #print(subjects)
   return (uri(subjects['values'][0][0]),uri(subjects['values'][0][1])) if len(subjects['values'])>0 else (None,None)

def mediaProperties(s,url,*names):
   service = app.config['UPDATE_SERVICE']
   expr = '<{0}> ?p ?s . ?s schema:contentUrl <{1}>'.format(s,url)
   for index,name in enumerate(names):
      expr += ';'
      expr += ' schema:'+name+' ?'+name
   q = SPARQL() \
         .start({ 'schema' : 'http://schema.org/'}) \
         .select(names) \
         .where(expr)
   #print(str(q))
   properties = service.query(q)
   return tuple(properties['values'][0]) if len(properties['values'])>0 else None

def subjectProperties(s,*names):
   service = app.config['UPDATE_SERVICE']
   expr = '<' + s + '>' if s[0]!='_' else s
   for index,name in enumerate(names):
      if index>0:
         expr += ';'
      expr += ' schema:'+name+' ?'+name
   q = SPARQL() \
         .start({ 'schema' : 'http://schema.org/'}) \
         .select(names) \
         .where(expr)
   #print(str(q))
   properties = service.query(q)
   return tuple(properties['values'][0]) if len(properties['values'])>0 else None

def jsonld_response(data):
   response = make_response(data)
   response.headers['Content-Type'] = "application/ld+json; charset=utf-8"
   return response

@app.route('/content/',methods=['GET','POST'])
def content():
   service = app.config['UPDATE_SERVICE']

   # GET returns a list of entities
   if request.method == 'GET':
      if 'reader' not in request.roles:
         abort(401)
      q = SPARQL() \
            .start({ 'schema' : 'http://schema.org/'}) \
            .select(['s','genre','type','name','headline','url','modified','published']) \
            .where('?s rdf:type ?type; schema:genre ?genre; schema:name ?name; schema:headline ?headline; schema:url ?url; schema:dateModified ?modified; . optional { ?s schema:datePublished ?published }'  )
      data = service.query(q)
      items = []
      for item in data['values']:
         ld = {
            '@context' : 'http://schema.org/',
            '@id' : uri(item[0]),
            '@type' : shorten(uri(item[2])),
            'genre' : literal(item[1]),
            'name' : literal(item[3]),
            'headline' : literal(item[4]),
            'url' : uri(item[5]),
            'dateModified' : literal(item[6])
         }
         if item[7] is not None:
            ld['datePublished'] = literal(item[7])
         items.append(ld)
      return jsonld_response(json.dumps(items))

   # POST creates an entity
   if request.method == 'POST':
      if 'writer' not in request.roles:
         abort(401)

      # Parse the incoming JSON-LD data
      force = request.headers['Content-Type'].startswith('application/ld+json')
      data = request.get_json(force=force)
      if data is None:
         abort(400)
      if 'name' not in data or \
         'genre' not in data or \
         '@type' not in data:
         abort(400)

      # Check to see if the name/genre already exists
      q = SPARQL() \
            .start({ 'schema' : 'http://schema.org/'}) \
            .select(['url']) \
            .where('?s schema:genre {}; schema:name {}; schema:url ?url'.format(value(data['genre']),value(data['name'])))
      existing = service.query(q)
      if len(existing['values'])>0:
         url = uri(existing['values'][0][0])
         id = url[url.rfind('/',0,-1)+1:-1]
         location = app.config['LOCATION_TEMPLATE'].format(**{'id' : id})
         return Response(status=409,headers={'Location' : location})
      if service.exists(ask({'genre' : data['genre'], 'name' : data['name']})):
         abort(409)

      subject = app.config['SUBJECT_TEMPLATE'].format(**data)
      id = str(uuid.uuid4())
      url = app.config['RESOURCE_TEMPLATE'].format(**{'id' : id})
      location = app.config['LOCATION_TEMPLATE'].format(**{'id' : id})
      data['url'] = url
      data['dateModified'] = now()

      turtle = TurtleSerializer()
      turtle.start()
      turtle.about(subject,data)
      service.put(turtle.end(),graph=url)
      return Response(status=201,headers={'Location' : location, 'Date-Modified' : data['dateModified']})

@app.route('/content/<id>/',methods=['GET','PUT','POST','DELETE'])
def content_item(id):
   service = app.config['UPDATE_SERVICE']
   url = app.config['RESOURCE_TEMPLATE'].format(**{'id' : id})
   subject = contentSubject(url)
   if subject == None:
      abort(404)
   if request.method == 'PUT':
      if 'writer' not in request.roles:
         abort(401)
      # Replace triples

      # Parse the incoming JSON-LD data
      force = request.headers['Content-Type'].startswith('application/ld+json')
      data = request.get_json(force=force)
      if data is None:
         abort(400)
      if 'name' not in data or \
         'genre' not in data or \
         '@type' not in data:
         abort(400)

      genre,name = subjectProperties(subject,'genre','name')

      # Check to see if the name/genre already exists
      if literal(genre)!=data['genre'] or literal(name)!=data['name']:
         if service.exists(ask({'genre' : data['genre'], 'name' : data['name']})):
            abort(409)

      subject = app.config['SUBJECT_TEMPLATE'].format(**data)
      data['url'] = url
      data['dateModified'] = now()

      turtle = TurtleSerializer()
      turtle.start()
      turtle.about(subject,data)

      # Update triples
      # TODO: transaction?
      service.deleteGraph(url)
      service.put(turtle.end(),graph=url)

      # Retrieve new representation
      quads = service.retrieve(graph=url)
      obj = toJSONLD(subject,quads)
      return jsonld_response(json.dumps(obj))

   if request.method == 'POST':
      if 'writer' not in request.roles:
         abort(401)
      force = request.headers['Content-Type'].startswith('application/ld+json')
      data = request.get_json(force=force)
      if data is None:
         abort(400)

      # Cannot rename subject with a post
      if 'name' in data or \
         'genre' in data:
         abort(400)

      data['dateModified'] = now()

      # Add triples
      turtle = TurtleSerializer()
      turtle.start()
      turtle.about(subject,data)
      service.post(turtle.end(),graph=url)

      # Retrieve new representation
      quads = service.retrieve(graph=url)
      obj = toJSONLD(subject,quads)
      return jsonld_response(json.dumps(obj))

   if request.method == 'GET':
      if 'reader' not in request.roles:
         abort(401)
      # Retrieve content
      quads = service.retrieve(graph=url)
      #print(quads)
      obj = toJSONLD(subject,quads)
      return jsonld_response(json.dumps(obj))

   if request.method == 'DELETE':
      if 'writer' not in request.roles:
         abort(401)
      service.deleteGraph(url)
      return Response(status=204)

@app.route('/content/<id>/<resource>',methods=['GET','PUT','DELETE'])
def content_item_resource(id,resource):
   return content_item_resource_property(id,resource,None)

@app.route('/content/<id>/<resource>;<property>',methods=['PUT'])
def content_item_resource_property(id,resource,property):
   service = app.config['UPDATE_SERVICE']
   url = app.config['RESOURCE_TEMPLATE'].format(**{'id' : id})
   resourceURL = url + resource
   parentSubject = contentSubject(url)
   if parentSubject == None:
      abort(404)
   subject,predicate = mediaSubject(parentSubject,resourceURL)
   #print(subject)
   if property is None:
      property = shorten(predicate)

   # allow PUT to create a resource
   if request.method == 'PUT':
      if 'writer' not in request.roles:
         abort(401)
      # Creating content
      contentType = request.headers['Content-Type']
      mediaObjectType = 'MediaObject'
      if contentType.find('image/')==0:
         mediaObjectType = 'ImageObject'
      elif contentType.find("video/")==0:
         mediaObjectType = 'VideoObject'
      if property is None:
         property = "associatedMedia"
      #print(request.headers)
      status = app.config['RESOURCE_SERVICE'].putResource(resourceURL,request.stream,content_type=contentType)
      if status>=200 and status<300:
         if subject is None:
            turtle = TurtleSerializer()
            turtle.start()
            turtle.about(parentSubject,{property : {"@type" : mediaObjectType, "contentUrl" : resourceURL, "fileFormat" : contentType, "name" : resource }})
            service.post(turtle.end(),graph=url)
            subject, *_ = mediaSubject(parentSubject,resourceURL)
         else:
            #print(subject)
            q = SPARQL() \
                  .start({ 'schema' : 'http://schema.org/'}) \
                  .withGraph(url) \
                  .delete('?media schema:fileFormat ?contentType') \
                  .insert('?media schema:fileFormat "{0}"'.format(contentType)) \
                  .where('?media schema:contentUrl <{0}>; schema:fileFormat ?contentType'.format(resourceURL) )
            service.update(q,graph=url)
      if status==204:
         status = 200
      if status>=200 and status<300:
         obj = {
            "@id" : subject,
            "@type" : mediaObjectType,
            "contentUrl" : resourceURL,
            "fileFormat" : contentType,
            "name" : resource
         }
         response = jsonld_response(json.dumps(obj))
         response.status_code = status
         return response
      else:
         abort(status)

   # when there is no subject for the associatedMedia, it does not exist
   if subject is None:
      abort(404)

   if request.method == 'GET':
      if 'reader' not in request.roles:
         abort(401)
      #print(subject)
      contentType, = mediaProperties(parentSubject,resourceURL,'fileFormat')
      #print(contentType)
      status,data,size = app.config['RESOURCE_SERVICE'].getResource(resourceURL)
      if status!=200:
         abort(status)
      response = Response(stream_with_context(data),mimetype=literal(contentType))
      response.headers['Content-Length'] = size
      return response

   if request.method == 'DELETE':
      if 'writer' not in request.roles:
         abort(401)
      # Retrieve content
      status = app.config['RESOURCE_SERVICE'].deleteResource(resourceURL)
      if status>=200 and status<300:
         # delete properties
         service.deleteProperties(parentSubject,object=subject,graph=url)
         service.deleteProperties(subject,graph=url)
      return Response(status=status)



from flask import render_template, redirect, session, request, url_for, escape, flash, g, make_response, stream_with_context, Response
from functools import wraps
import json
import requests
from .app import app
from .forms import LoginForm
from . import model

@app.route('/')
def index():
   return render_template('home.html',
                           title='Duckpond Content Editor',
                           username=session['username'])
@app.route('/config.json')
def config():
   config = app.config.get('EDITOR_CONFIG')
   if config is None:
      config = { 'wrap-header' : ''}
   response = make_response(json.dumps(config))
   response.headers['Content-Type'] = "application/json; charset=utf-8"
   return response


@app.route('/login', methods=['GET', 'POST'])
def login():
   if 'username' in session:
      return redirect('/')
   form = LoginForm()
   failed = False
   if form.validate_on_submit():
      username = form.username.data
      password = form.password.data
      if app.config['AUTH_SERVICE'].authenticate(username,password):
         g.user = app.config['AUTH_SERVICE'].getUser(username)
      try:
         if g.user is not None:
            flash('Login for {} successful.'.format(username))
            session['username'] = username
            print("rediect="+form.redirectonlogin.data)
            if form.redirectonlogin.data=='true':
               return redirect('/')
            else:
               return render_template('login-successful.html',
                                      title='Duckpond Content Editor',
                                      username=session['username'])
         else:
            flash('Login for {} failed.'.format(username))
            failed = True
      except AttributeError:
         flash('Login for {} failed.'.format(username))
         failed = True

   compact = request.args.get('compact')
   requestRedirect = request.args.get('redirect')
   form.redirectonlogin.data = requestRedirect if requestRedirect is not None else 'true'

   return render_template('login.html' if compact is None or compact=='false' else 'login-compact.html',
                           title='Sign In',
                           failed=failed,
                           form=form)

@app.route('/logout')
def logout():
    # remove the username from the session if it's there
    session.clear()
    return redirect(url_for('index'))

@app.route('/proxy/<name>/<path:path>')
def home(name,path):
   proxies = app.config.get('PROXIES')
   if proxies is None:
      abort(404)
   base = proxies.get(name)
   if base is None:
      abort(404)
   req = requests.get(base + path, stream = True)
   return Response(stream_with_context(req.iter_content()), content_type = req.headers['content-type'])



from flask import Blueprint, request, render_template, abort, send_file, redirect, current_app
import logging,collections

from duckpond.data import Pond,uripath
from duckpond.data.pond import Facet

views = Blueprint('duckpond_blog_views',__name__,template_folder='templates')

logger = logging.getLogger('webapp')

def getPond(config):
   user = config.get('USER')
   password = config.get('PASSWORD')

   data = Pond(config['SPARQL_SERVICE'],cache=config['CACHE'],facets=[ Facet('resource','schema:isBasedOnUrl','xsd:anyURI')])
   data.authenticate(user if user is not None else 'anonymous', password if password is not None else '')
   return data

def getJournalCriteria(config):
   journalGenre = config.get('JOURNAL_GENRE')
   if journalGenre is None:
      journalGenre = 'blog'
   return {'schema:genre' : journalGenre}


def getTemplateOptions(config):
   options = config.get('TEMPLATE_OPTIONS')
   return options if options is not None else {
      'title' : 'My Journal'
   }

def entry(e):
   if e is None:
      return None
   else:
      return {'uri':uripath(e[0]),'subject':e[0],'title':e[1],'summary':e[2],'date':e[3],'time':e[4],'basedOnUrl':e[5]}

def siteURL(config,request):
   u = current_app.config.get('SITE_URL')
   return u if u is not None else request.url_root[0:-1]

def renderEntry(data,basedOnUrl,entry=None,preceding=None,following=None,base=None,path=None):
   try:
      logger.debug(basedOnUrl)
      content = data.getResourceText(basedOnUrl)
      topics = [(key,value) for key,value in data.getCategoryCount().items()]
      sortedTopics = sorted(topics,key=lambda x : str.lower(x[0]))
      return render_template('base.html', siteURL=siteURL(current_app.config,request), path=path, options=getTemplateOptions(current_app.config), entry=entry,entryContent=content,preceding=preceding,following=following,keywords=sorted(data.getEntityCategories(entry['subject']),key=str.lower),topics=sortedTopics,base=base)
   except FileNotFoundError:
      abort(404)

def renderKeyword(data,keyword,related,entries):
   try:
      topics = [(key,value) for key,value in data.getCategoryCount().items()]
      sortedTopics = sorted(topics,key=lambda x : str.lower(x[0]))
      return render_template('keyword.html', entry=None, siteURL=siteURL(current_app.config,request), path=request.path, options=getTemplateOptions(current_app.config), keyword=keyword,related=sorted(related,key=str.lower),entries=entries,topics=sortedTopics)
   except FileNotFoundError:
      abort(404)

@views.route('/')
def index():

   data = getPond(current_app.config)

   entryInfo = data.currentEntity(criteria=getJournalCriteria(current_app.config))
   if entryInfo is None:
      abort(404)

   subject,_,_,date,time,*_ = entryInfo
   dateTime = date + "T" + time
   resource = data.getEntityResource(subject)
   following = data.relatedEntityByOrder(dateTime,previous=True)
   path = "/journal/entry/{}T{}/".format(date,time)
   return renderEntry(data,resource[0],entry=entry(entryInfo),following=entry(following) if following is not None and len(following)>0 else None,base=path,path=path)

@views.route('/journal/entry/<date>T<time>/')
def entryByTime(date,time):

   data = getPond(current_app.config)

   dateTime = date + "T" + time
   entryInfo = data.entityByOrder(dateTime)
   if entryInfo is None:
      abort(404)

   following = data.relatedEntityByOrder(dateTime,previous=True)
   preceding = data.relatedEntityByOrder(dateTime,previous=False)
   resource = data.getEntityResource(entryInfo[0])

   return renderEntry(data,resource[0],entry=entry(entryInfo),following=entry(following),preceding=entry(preceding),path=request.path)

@views.route('/journal/entry/<date>T<time>/<path:path>')
def entryMedia(date,time,path):

   data = getPond(current_app.config)

   uri = app.config['CACHE']['base'] + date + '/' + path
   resource = data.getResource(uri)
   if resource[0] == Pond.ResourceType.uri:
      return redirect(resource[1], code=resource[2])
   elif resource[0] == Pond.ResourceType.stream:
      return Response(stream_with_context(resource[1]), content_type = resource[2])
   else:
      return send_file(resource[1])

@views.route('/rel/keyword/<keyword>')
def relKeyword(keyword):

   data = getPond(current_app.config)

   entries = data.getEntitiesByCategory(keyword)
   allKeywords = []
   entries = list(map(lambda e:entry(e),entries))
   for relatedEntry in entries:
      relatedEntry['keywords'] = sorted(data.getEntityCategories(relatedEntry['subject']))
      allKeywords.extend(relatedEntry['keywords'])
   return renderKeyword(data,keyword,set(allKeywords),entries)

@views.errorhandler(404)
def page_not_found(error):
   return render_template('error.html', siteURL=siteURL if siteURL is not None else request.url_root[0:-1], path=request.path, options=templateOptions, entry=None, error="I'm sorry.  I can't find that page.")



from flask import Flask, request, g, session, redirect, abort, Response
from datetime import datetime
from functools import wraps
import base64

app = Flask(__name__)
app.config.from_envvar('WEB_CONF')

def authenticate(f):
   @wraps(f)
   def wrapper(*args, **kwargs):
      v = request.headers.get('authorization')
      authenticated = False
      if v is not None and v.find('Basic ')==0:
         s = str(base64.b64decode(v[6:]),'utf-8')
         username = s[0:s.find(':')]
         password = s[s.find(':')+1:]
         info = app.config['AUTH_SERVICE'].authenticate(username,password);
         if info is not None:
            request.roles = info.get('roles')
            authenticated = True
      if v is not None and v.find('Bearer ')==0:
         roles = app.config['AUTH_SERVICE'].validateToken(v[7:])
         if roles is not None:
            authenticated = True
            request.roles = roles
      if authenticated:
         return f(*args, **kwargs)
      else:
         return Response(status=401, headers={'WWW-Authenticate': 'Basic realm="service"'})
   return wrapper

@app.before_request
@authenticate
def before_request():
   pass



from flask import Flask
import logging, logging.config, os

# TODO: total hack to get the configuration of the templates folder
f = os.environ.get('WEB_CONF')
template_folder = 'templates'
if f is not None:
   d = {}
   with open(f) as config_file:
       exec(compile(config_file.read(), f, 'exec'), d)

   template_folder = d.get('TEMPLATE_FOLDER')
   if template_folder is None:
      template_folder = 'templates'
# END

app = Flask(__name__,template_folder=template_folder)
app.config.from_envvar('WEB_CONF')
logLevel = app.config.get('LOGLEVEL')
if logLevel is not None:
   logging.basicConfig(level=logLevel)

logConfig = app.config.get("LOGCONFIG")
if logConfig is not None:
   logging.config.dictConfig(logConfig)



import json, os, io
from flask import make_response, request, Response, stream_with_context, abort
from .app import app
from . import model

def jsonld_response(data):
   response = make_response(data)
   response.headers['Content-Type'] = "application/ld+json; charset=utf-8"
   return response

@app.route('/data/content/',methods=['GET','POST'])
def content():
   if request.method == 'GET':
      works = model.getContentList()
      return jsonld_response(json.dumps(works))

   if request.method == 'POST':
      # Parse the incoming JSON-LD data
      force = request.headers['Content-Type'].startswith('application/ld+json')
      data = request.get_json(force=force)
      if data is None:
         abort(400)
      if 'name' not in data or \
         'genre' not in data or \
         'headline' not in data or \
         '@type' not in data:
         abort(400)
      status,url,modified = model.createContent(data['@type'],data['genre'],data['name'],data['headline'])
      return Response(status=status,headers=({'Location' : url, 'Date-Modified' : modified} if status==201 else {}))

@app.route('/data/content/<id>/',methods=['GET','PUT','POST','DELETE'])
def content_item(id):

   if request.method == 'GET':
      content = model.getContent(id)
      return jsonld_response(json.dumps(content))

   if request.method == 'POST':
      abort(400)

   if request.method == 'PUT':
      force = request.headers['Content-Type'].startswith('application/ld+json')
      data = request.get_json(force=force)
      if data is None:
         abort(400)
      status_code,data,contentType = model.updateContent(id,data);
      if status_code==200:
         return Response(stream_with_context(data),content_type = contentType)
      else:
         abort(status_code)

   if request.method == 'DELETE':
      status = model.deleteContent(id)
      return Response(status=status)

@app.route('/data/content/<id>/<resource>',methods=['GET','PUT','DELETE'])
def content_item_resource(id,resource):
   if request.method == 'GET':
      wrap = request.args.get('wrap')
      status_code,data,contentType = model.getContentResource(id,resource);
      if status_code==200:
         if contentType.startswith("text/html") and wrap is not None:
            blob = io.BytesIO()
            for chunk in data:
               blob.write(chunk)
            content = blob.getvalue().decode("utf-8").strip()
            if not content.startswith('<!DOCTYPE'):
               editorConfig = app.config.get('EDITOR_CONFIG')
               header = ''
               bodyStart = ''
               bodyEnd = ''
               if editorConfig is not None and wrap=='preview':
                  wheader = editorConfig.get('wrap-header')
                  pheader = editorConfig.get('preview-wrap-header')
                  if pheader is not None:
                     header = pheader
                  elif wheader is not None:
                     header = wheader
                  wbody = editorConfig.get('wrap-body')
                  pbody = editorConfig.get('preview-body-main')
                  if pbody is not None:
                     bodyStart = pbody[0]
                     bodyEnd = pbody[1]
                  elif wbody is not None:
                     bodyStart = wbody[0]
                     bodyEnd = wbody[1]
               elif editorConfig is not None and wrap=='formatted':
                  wheader = editorConfig.get('wrap-header')
                  if wheader is not None:
                     header = wheader
                  wbody = editorConfig.get('wrap-body')
                  if wbody is not None:
                     bodyStart = wbody[0]
                     bodyEnd = wbody[1]
               content = """
<!DOCTYPE html>
<html>
<head><title>""" + resource + '</title>' + header + """
</head>
<body>
""" + bodyStart + content + bodyEnd + '</body></html>'
            return Response(stream_with_context(content),content_type = contentType)
         else:
            return Response(stream_with_context(data),content_type = contentType)
      else:
         abort(status_code)
   if request.method == 'PUT':
      status_code,data,contentType = model.updateContentResource(id,resource,request.headers['Content-Type'],request.stream);
      if status_code==200 or status_code==201:
         return Response(stream_with_context(data),status=status_code,content_type = contentType)
      else:
         return Response(status=status)

   if request.method == 'DELETE':
      status = model.deleteContentResource(id,resource)
      return Response(status=status)


@app.route('/data/content/<id>/upload/<property>',methods=['POST'])
def content_item_resource_upload(id,property):
   #print(request.headers['Content-Type'])
   #print(request.files)
   file = request.files['file']
   #print(file.filename)
   #print(file.content_type)
   #print(file.content_length)
   uploadContentType = file.content_type
   if file.content_type.startswith("text/") and file.content_type.find("charset=")<0:
      uploadContentType = file.content_type+"; charset=UTF-8"
   uploadDir = app.config['UPLOAD_STAGING'] if 'UPLOAD_STAGING' in app.config else 'tmp'
   os.makedirs(uploadDir,exist_ok=True)
   staged = os.path.join(uploadDir, file.filename)
   file.save(staged)
   status = 500
   responseJSON = None
   contentType = None
   with open(staged,"rb") as data:
      status,responseJSON,contentType = model.uploadContentResource(id,property,file.filename,uploadContentType,os.path.getsize(staged),data)
   os.unlink(staged)
   if status==200 or status==201:
      return Response(stream_with_context(responseJSON),status=status,content_type = contentType)
   else:
      return Response(status=status)




import sys,io,os,argparse

import json,requests,urllib

argparser = argparse.ArgumentParser(description='Triple loader')
argparser.add_argument('dir',nargs=1,help='The directory to process.')
argparser.add_argument('uri',nargs=1,help='The repository URI.')
argparser.add_argument('graph',nargs='?',help='The graph name')
argparser.add_argument('-u',help='The user',dest='user')
argparser.add_argument('-p',help='The password',dest='password')
args = argparser.parse_args()
inDir = args.dir[0]
repository = args.uri[0]

auth = requests.auth.HTTPBasicAuth(args.user, args.password)

def loadTriples(targetFile):

   print(targetFile)
   f = open(targetFile,"r",encoding='utf-8')
   data = f.read()
   f.close()

   uri = repository
   if args.graph is not None and len(args.graph)>0:
      uri = uri + "?context=" + urllib.parse.quote('<'+args.graph+'>')
   req = requests.post(uri,data=data.encode('utf-8'),headers={'content-type':'text/turtle; charset=utf-8'},auth=auth)
   if (req.status_code<200 or req.status_code>=300):
      raise IOError('Cannot post data to uri <{}>, status={}'.format(uri,req.status_code))

for file in [f for f in os.listdir(inDir) if f.endswith('.ttl') and os.path.isfile(inDir + '/' + f)]:
   targetFile = inDir + '/' + file

   loadTriples(targetFile)

dirs = [d for d in os.listdir(inDir) if not(d[0]=='.') and os.path.isdir(inDir + '/' + d)]

for dir in dirs:
   sourceDir = inDir + '/' + dir

   files = [f for f in os.listdir(sourceDir) if f.endswith('.ttl') and os.path.isfile(sourceDir + '/' + f)]
   for file in files:
      targetFile = sourceDir + '/' + file

      loadTriples(targetFile)



import shutil, os
from io import BufferedReader
from tempfile import TemporaryFile
from urllib.parse import urlparse
import boto3, botocore

class FileResourceService:
   def __init__(self,base=None,directory='.',cleanup=True):
      self.base = base
      self.directory = directory
      self.separator = '/' if self.directory[-1]!='/' else ''
      self.cleanup = True

   def resolveLocation(self,url):
      path = None
      if url.find(self.base)==0:
         path = url[len(self.base):]
      else:
         colon = url.find(':')
         if url[colon:colon+3]=='://':
            slash = url.find('/',colon+3)
            path = url[slash+1:]
         else:
            raise ValueError('The URL is not generic: '+url)
      return self.directory + self.separator + path

   def getResource(self,url):
      location = self.resolveLocation(url)
      if os.path.isfile(location):
         return (200,open(location,'rb'),os.path.getsize(location))
      else:
         return (404,None,None)

   def putResource(self,url,stream,content_type=None):
      location = self.resolveLocation(url)
      print(location)
      last = location.rfind('/')
      dirpath = location[0:last]
      os.makedirs(dirpath,exist_ok=True)
      exists = os.path.isfile(location)
      with open(location,'wb') as fdest:
         shutil.copyfileobj(stream,fdest)
      return 204 if exists else 201

   def deleteResource(self,url):
      location = self.resolveLocation(url)
      if not os.path.isfile(location):
         return 404
      os.remove(location)

      if self.cleanup:
         # Try to remove empty content directories
         try:
            last = location.rfind('/')
            dirpath = location[0:last]
            os.rmdir(dirpath)
         except OSError:
            # An exception means the directory is not empty and that's okay!
            pass
      return 200

class S3ResourceService:
   def __init__(self,access_key_id,secret_access_key,bucket,bucket_domains={},tmpdir=None):
      self.s3 = boto3.client('s3',aws_access_key_id=access_key_id,aws_secret_access_key=secret_access_key)
      self.bucket_name = bucket
      self.bucket_domains = bucket_domains
      self.tmpdir = tmpdir
      self.readSize = 32*1024

   def parseLocation(self,url):
      spec = urlparse(url)
      colon = spec.netloc.find(':')
      return (spec.netloc[0:colon] if colon>0 else spec.netloc,spec.path[1:])

   def getBucket(self,domain):
      bucket = self.bucket_domains.get(domain)
      return bucket if bucket is not None else self.bucket_name

   def getResource(self,url):
      domain,path = self.parseLocation(url)
      try:
         response = self.s3.get_object(Bucket=self.getBucket(domain),Key=path)
         def readBody():
            while True:
               data = response['Body'].read(self.readSize)
               if len(data)>0:
                  yield data
               else:
                  response['Body'].close()
                  break
         return (200,readBody(),response['ContentLength'])
      except botocore.exceptions.ClientError as e:
         if e.response['Error']['Code'] == "404":
            return (404,None,None)
         else:
            raise e

   def putResource(self,url,stream,content_type=None):
      domain,path = self.parseLocation(url)
      creating = False
      try:
         self.s3.head_object(Bucket=self.getBucket(domain),Key=path)
      except botocore.exceptions.ClientError as e:
         if e.response['Error']['Code'] == "404":
            creating = True
         else:
            raise e
      with TemporaryFile(dir=self.tmpdir) as cached:
         shutil.copyfileobj(stream,cached)
         cached.seek(0)
         self.s3.put_object(Bucket=self.getBucket(domain),Body=cached,Key=path,ContentType=content_type)
      return 201 if creating else 204

   def deleteResource(self,url):
      domain,path = self.parseLocation(url)
      try:
         self.s3.delete_object(Bucket=self.getBucket(domain),Key=path)
      except botocore.exceptions.ClientError as e:
         if e.response['Error']['Code'] != "404":
            raise e
      return 204




from duckpond.apps.blog.app import app
from duckpond.apps.blog.assets import assets
from duckpond.apps.blog.docs import docs
from duckpond.apps.blog.views import views
app.register_blueprint(assets)
app.register_blueprint(docs)
app.register_blueprint(views)

if __name__ == '__main__':
    app.run()




from duckpond.apps.editor import app
from duckpond.apps.editor.assets import *
from duckpond.apps.editor.views import *
from duckpond.apps.editor.api import *
from duckpond.apps.editor.model import *

if __name__ == '__main__':
    app.run()



import CommonMark,re,argparse

import sys,io,os,shutil

from html import escape

META_RE = re.compile(r'^[ ]{0,3}(?P<key>[A-Za-z0-9_-]+):\s*(?P<value>.*)')
META_MORE_RE = re.compile(r'^[ ]{4,}(?P<value>.*)')

class ArticleConverter:
   def __init__(self,weburi,entryuri):
      self.weburi = weburi
      self.entryuri = [ entryuri ]
      self.vocab = 'http://schema.org'
      self.typeof = 'BlogPosting'

   def enter(self,suffix):
      self.entryuri.append(self.entryuri[-1] + suffix)

   def leave(self):
      return self.entryuri.pop()
      
   def toArticle(self,base,md,html,triples):
      glob = io.StringIO()
      metadata = {}
      hasMetadata = True
      hasTitle = False
      line = md.readline()
      while hasMetadata:
         m = META_RE.match(line)
         if (not(m)):
            glob.write(line)
            if not hasTitle and line[0:2]=='# ':
               hasTitle = True
            glob.write('\n')
            hasMetadata = False
         else:
            key = m.group('key').lower().strip()
            value = m.group('value').strip()
            try:
               metadata[key].append(value)
            except KeyError:
               metadata[key] = [value]
            continued = True
            while continued:
               line = md.readline()
               more = META_MORE_RE.match(line)
               if (not(more)):
                  continued = False
               else:
                  metadata[key].append(more.group('value').strip())

      for line in md:
         if not hasTitle and line[0:2]=='# ':
            hasTitle = True
         glob.write(line)

      self.toHTML(glob.getvalue(),metadata,html,generateTitle=not hasTitle)
      self.toTurtle(base,metadata,triples)

   def toHTML(self,md,metadata,html,generateTitle=False):

      uri = self.weburi + metadata['published'][0]

      print('<article xmlns="http://www.w3.org/1999/xhtml" vocab="{}" typeof="{}" resource="{}">'.format(self.vocab,self.typeof,uri),file=html)
   
      print('<script type="application/json+ld">',file=html)
      print('{\n"@context" : "http://schema.org/",',file=html)
      print('"@id" : "{}",'.format(uri),file=html)
      print('"headline" : "{}",'.format(metadata['title'][0]),file=html)
      print('"datePublished" : "{}",'.format(metadata['published'][0]),file=html)
      print('"dateModified" : "{}",'.format(metadata['updated'][0]),file=html)
      if ("keywords" in metadata):
         print('"keywords" : [{}],'.format(','.join(['"' + m + '"' for m in metadata['keywords']])),file=html)
      html.write('"author" : [ ')
      for index,author in enumerate(metadata['author']):
         if (index>0):
            html.write(', ')
         html.write('{{ "@type" : "Person", "name" : "{}" }}'.format(author))
      html.write(']\n')
      print('}',file=html)
      print('</script>',file=html)

      if generateTitle:
         print("<h1>{}</h1>".format(escape(metadata['title'][0],quote=False)),file=html)

      print(CommonMark.commonmark(md),file=html)

      print('</article>',file=html)

   def toTurtle(self,base,metadata,triples):
      
      uri = self.weburi + metadata['published'][0]
      basedOn = self.entryuri[-1] + base + '.html'
      
      print('@base <{}> .'.format(uri),file=triples)
      print('@prefix schema: <http://schema.org/> .',file=triples)
      print('<>',file=triples)
      print('   a schema:BlogPosting ;',file=triples)
      print('   schema:headline "{}" ;'.format(metadata['title'][0]),file=triples)
      print('   schema:datePublished "{}" ;'.format(metadata['published'][0]),file=triples)
      print('   schema:dateModified "{}" ;'.format(metadata['updated'][0]),file=triples)
      print('   schema:isBasedOnUrl <{}> ;'.format(basedOn),file=triples)
      if ("keywords" in metadata):
         print('   schema:keywords {} ;'.format(','.join(['"' + m + '"' for m in metadata['keywords']])),file=triples)
      triples.write('   schema:author ')
      for index,author in enumerate(metadata['author']):
         if (index>0):
            triples.write(', ')
         triples.write('[ a schema:Person; schema:name "{}" ]'.format(author))
      triples.write(' .\n')

if __name__ == "__main__":
   argparser = argparse.ArgumentParser(prog=sys.modules[__name__].__spec__.name,description='Article HTML and Turtle Generator')
   argparser.add_argument('-f',action='store_true',help='Forces all the files to be regenerated.',dest='force')
   argparser.add_argument('-o',nargs='?',help='The output directory',dest='outdir')
   argparser.add_argument('-w',nargs='?',help='The web uri',dest='weburi',default='http://www.milowski.com/journal/entry/')
   argparser.add_argument('-e',nargs='?',help='The entry uri directory',dest='entryuri',default='http://alexmilowski.github.io/milowski-journal/')
   argparser.add_argument('--immediate',action='store_true',help='Only process the immediate director (no subdirectories)',dest='immediate')
   argparser.add_argument('dir',nargs=1,help='The directory to process.')
   args = argparser.parse_args()
   inDir = args.dir[0]
   outDir = args.outdir if (args.outdir!=None) else inDir
   dirs = [inDir] if args.immediate else [d for d in os.listdir(inDir) if not(d[0]=='.') and os.path.isdir(inDir + '/' + d)]

   converter = ArticleConverter(args.weburi,args.entryuri)

   for dir in dirs:
   
      sourceDir = inDir + '/' + dir if inDir!='.' else dir
      targetDir = outDir + '/' + dir if outDir!='.' else dir
   
      if (not(os.path.exists(targetDir))):
         os.makedirs(targetDir)

      converter.enter(dir+'/')
   
      files = [f for f in os.listdir(sourceDir) if f.endswith('.md') and os.path.isfile(sourceDir + '/' + f)]
   
      for file in files:
      
         targetFile = sourceDir + '/' + file
         base = file.rsplit('.md',1)[0]
         htmlFile = targetDir + '/' + base + ".html"
         turtleFile = targetDir + '/' + base + ".ttl"
      
         updatedNeeded = args.force or not(os.path.exists(htmlFile)) or not(os.path.exists(turtleFile))
         if (not(updatedNeeded)):
            btime = os.path.getmtime(targetFile)
            updatedNeeded = btime > os.path.getmtime(htmlFile) or btime > os.path.getmtime(turtleFile)
         
         if (not(updatedNeeded)):
            continue
      
         print(file + " → " + htmlFile + ", " + turtleFile)
      
         md = open(targetFile,"r")
         html = open(htmlFile,"w")
         turtle = open(turtleFile,"w")
      
         converter.toArticle(base,md,html,turtle)
      
         md.close()
         html.close()
         turtle.close()

      work = [f for f in os.listdir(sourceDir) if (not f.endswith('.md'))]
      for file in work:
         sourceFile = sourceDir + '/' + file
         targetFile = targetDir + '/' + file
         if os.path.isfile(sourceFile):
            copyNeeded = args.force or not(os.path.exists(targetFile)) or os.path.getmtime(sourceFile) > os.path.getmtime(targetFile)
         
            if copyNeeded:
               print(file + " → " + targetFile)
               shutil.copyfile(sourceFile,targetFile)
         elif os.path.isdir(sourceFile):
            if os.path.exists(targetFile):
               print('sync tree ' + sourceFile)
               work += [sourceFile + '/' + f for f in os.listdir(sourceFile)]
            else:
               print("copy tree " + sourceFile + " → " + targetFile)
               shutil.copytree(sourceFile,targetFile)
   
      converter.leave()




from flask import send_from_directory
from .app import app

@app.route('/assets/<path:path>')
def send_asset(path):
   dir = __file__[:__file__.rfind('/')] + '/assets/'
   return send_from_directory(dir, path)



import requests
import json
import shutil
from tempfile import TemporaryFile
from flask import session, abort
from .app import app

def getAuth():
   user = app.config['AUTH_SERVICE'].getUser(session['username'])
   return requests.auth.HTTPBasicAuth(user['username'], user['password'])

def getContentList():
   url = app.config['SERVICE'] + 'content/'
   req = requests.get(url,auth=getAuth(),headers={'Connection' : 'close'})
   if req.status_code!=200:
      raise IOError('Cannot get content list from service, status {status}'.format(**{'status':req.status_code}))
   return json.loads(req.text)

def createContent(typeName,genre,name,headline):
   url = app.config['SERVICE'] + 'content/'
   user = app.config['AUTH_SERVICE'].getUser(session['username'])
   headers = {'Content-Type' : 'application/ld+json', 'Connection' : 'close'}
   data = {
      '@context' : 'http://schema.org/',
      '@type' : typeName,
      'genre' : genre,
      'name' : name,
      'headline' : headline
   }
   response = requests.post(url,auth=getAuth(),data=json.dumps(data),headers=headers)
   if response.status_code!=201:
      return (response.status_code,None,None)
   else:
      return (response.status_code,response.headers['Location'],response.headers['Date-Modified'])

def deleteContent(id):
   url = app.config['SERVICE'] + 'content/' + id + '/'
   response = requests.delete(url,auth=getAuth(),headers={'Connection' : 'close'})
   return response.status_code

def getContent(id):
   url = app.config['SERVICE'] + 'content/' + id + '/'
   response = requests.get(url,auth=getAuth(),headers={'Connection' : 'close'})
   if response.status_code==200:
      return json.loads(response.text)
   else:
      raise IOError('Cannot get content {id} from service, status {status}'.format(**{'status':response.status_code,'id':id}))

def updateContent(id,data):
   url = app.config['SERVICE'] + 'content/' + id + '/'
   headers = {'Content-Type' : 'application/ld+json', 'Connection' : 'close'}
   response = requests.put(url,auth=getAuth(),data=json.dumps(data),headers=headers)
   return (response.status_code,response.iter_content(chunk_size=10*1024),response.headers.get('content-type'))

def getContentResource(id,name):
   url = app.config['SERVICE'] + 'content/' + id + '/' + name
   response = requests.get(url,auth=getAuth(),stream = True,headers={'Connection' : 'close'})
   return (response.status_code,response.iter_content(chunk_size=10*1024),response.headers.get('content-type'))

def updateContentResource(id,name,content_type,data):
   url = app.config['SERVICE'] + 'content/' + id + '/' + name
   with TemporaryFile() as cached:
      shutil.copyfileobj(data,cached)
      size = cached.tell()
      cached.seek(0)
      headers = {'Content-Type' : content_type, 'Content-Length' : str(size), 'Connection' : 'close'}
      response = requests.put(url,auth=getAuth(),data=cached,headers=headers)
      return (response.status_code,response.iter_content(chunk_size=10*1024),response.headers.get('content-type'))

def uploadContentResource(id,property,name,content_type,content_length,data):
   url = app.config['SERVICE'] + 'content/' + id + '/' + name + ";" + property
   headers = {'Content-Type' : content_type, 'Content-Length' : str(content_length), 'Connection' : 'close'}
   response = requests.put(url,auth=getAuth(),data=data,headers=headers)
   return (response.status_code,response.iter_content(chunk_size=10*1024),response.headers.get('content-type'))

def deleteContentResource(id,name):
   url = app.config['SERVICE'] + 'content/' + id + '/' + name
   response = requests.delete(url,auth=getAuth(),headers={'Connection' : 'close'})
   return response.status_code



"""A semantic pond for content delivery.

See:

https://github.com/alexmilowski/duckpond
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

long_description = """
A [semantic data lake](http://cacm.acm.org/news/200095-the-data-lake-concept-is-maturing/fulltext) can
be thought of a set of resources annotated with semantics (i.e., triples or quads).  In the context of big data,
a "data lake" is a massive repository of data, structured or unstructured, of which we gain knowledge of
its contents by examining the semantic graphs derived from their contents.

That's a big idea.  This project is about a *semantic data lake* at a much smaller
scale: a pond. I also like ducks and so it is a *duck pond*.

We can paddle around in our content, just like a duck, and harvest a bit knowledge
to derive local value.  In this case, we use the semantic pond to understand the
content, its inter-relations, ordering, and other such content relations.  From
that, we can derive a useful presentation on the Web.
"""

setup(
    name='duckpond',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.9.1',

    description='A semantic pond for content delivery',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/alexmilowski/duckpond',

    # Author details
    author='Alex Miłowski',
    author_email='alex@milowski.com',

    # Choose your license
    license='Apache 2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    # What does your project relate to?
    keywords='sparql semantics cms',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['commonmark', 'requests', 'Flask', 'Flask-Session', 'Flask-WTF', 'WTForms', 'pytz', 'boto3', 'botocore'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
    },

    include_package_data=True,

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
    },
)



from flask import Blueprint
from flask import send_from_directory, current_app

assets = Blueprint('duckpond_blog_assets',__name__)
@assets.route('/assets/<path:path>')
def send_asset(path):
   dir = current_app.config.get('ASSETS')
   if dir is None:
      dir = __file__[:__file__.rfind('/')] + '/assets/'
   return send_from_directory(dir, path)



import io

class SPARQL:

   def __init__(self):
      self.statement = io.StringIO()

   def start(self,prefixes):
      for key in prefixes:
         self.statement.write('prefix {}: <{}>\n'.format(key,prefixes[key]))
      return self

   def select(self,names):
      self.statement.write('select ')
      for name in names:
         if name[0]=='(':
            self.statement.write(name)
         else:
            self.statement.write('?')
            self.statement.write(name)
         self.statement.write(' ')
      self.statement.write('\n')
      return self

   def fromGraphs(self,graphs):
      for graph in graphs:
         self.statement.write('from <{0}>\n'.format(graph))
      return self

   def fromNameGraphs(self,graphs):
      for graph in graphs:
         self.statement.write('from named <{0}>\n'.format(graph))
      return self

   def withGraph(self,graph):
      self.statement.write("with <{0}>\n".format(graph))
      return self

   def where(self,*expressions):
      self.statement.write('where {')
      for expression in expressions:
         self.statement.write(str(expression))
         self.statement.write('\n')
      self.statement.write('}\n')
      return self

   def insert(self,*expressions):
      self.statement.write('insert {')
      for expression in expressions:
         self.statement.write(str(expression))
         self.statement.write('\n')
      self.statement.write('}\n')
      return self

   def delete(self,*expressions):
      self.statement.write('delete {')
      for expression in expressions:
         self.statement.write(str(expression))
         self.statement.write('\n')
      self.statement.write('}\n')
      return self

   def anyGraph(self,*expressions):
      self.statement.write('graph ?g {')
      for expression in expressions:
         self.statement.write(str(expression))
         self.statement.write('\n')
      self.statement.write('}\n')
      return self

   def forGraph(self,graph,*expressions):
      self.statement.write('graph <{0}> {'.format(graph))
      for expression in expressions:
         self.statement.write(str(expression))
         self.statement.write('\n')
      self.statement.write('}\n')
      return self

   def newline(self):
      self.statement.write('\n')
      return self;

   def orderBy(self,*expressions):
      self.statement.write('order by ')
      first = True
      for expression in expressions:
         if not first:
            self.statement.write(',')
         self.statement.write(str(expression))
      self.statement.write('\n')
      return self

   def groupBy(self,expression):
      self.statement.write('group by ')
      self.statement.write(expression)
      self.statement.write('\n')
      return self

   def __str__(self):
      return self.statement.getvalue()



import requests
import json,io,os
import logging
from enum import Enum
import urllib
from .sparql import SPARQL

def strip(value):
   if value is None:
      return value
   elif value[0]=='"' and value[-1]=='"':
      value = value[1:-1]
      if value[0]=='<' and value[-1]=='>':
         return value[1:-1]
      else:
         return value
   elif value[0]=='<':
      return value[1:-1]
   else:
      return value

def literal(value):
   value = strip(value)
   return value.replace('\\"','"')

class Facet:
   def __init__(self,id,name,type = 'xsd:string'):
      self.id = id
      self.name = name
      self.type = type

   def __str__(self):
      return self.name

   def toLiteral(self,value):
      if self.type=='xsd:anyURI':
         return '<' + value + '>'
      else:
         return '"' + value + '"'

orgSchema = [
   Facet('type','schema:BlogPosting',None),
   Facet('order','schema:datePublished','xsd:dateTime'),
   Facet('title','schema:headline'),
   Facet('summary','schema:description'),
   Facet('name','schema:name'),
   Facet('genre','schema:genre'),
   Facet('resource','schema:hasPart','schema:MediaObject'),
   Facet('category','schema:keywords'),
   Facet('contentUrl','schema:contentUrl'),
   Facet('fileFormat','schema:fileFormat'),
   Facet('associatedMedia','schema:associatedMedia','schema:MediaObject'),

   Facet('Article','schema:Article'),
   Facet('MediaObject','schema:MediaObject'),
   Facet('ImageObject','schema:ImageObject'),
   Facet('ImageGallery','schema:ImageGallery')
]

def criteriaConditions(pond,criteria,subjectVar='s'):
   conditions = io.StringIO()
   def addLiteralCriteria(property,value):
      conditions.write('\n?{0} {1} "{2}" .'.format(subjectVar,property,value))
   def addURICriteria(property,value):
      conditions.write('\n?{0} {1} <{2}> .'.format(subjectVar,property,value))
   for property in criteria.keys():
      value = criteria[property]
      if property.find(':')<0:
         property = 'schema:' + property
      if type(value)==dict:
         literal = value.get('@value')
         typename = value.get('@type')
         if literal is not None:
            if typename=='xsd:anyURI' or pond.isURIProperty(property):
               addURICriteria(property,literal)
            else:
               addLiteralCriteria(property,literal)
      elif pond.isURIProperty(property):
         addURICriteria(property,value)
      else:
         addLiteralCriteria(property,value)
   return conditions.getvalue()

class Pond:

   class AuthMethod(Enum):
      basic = 1
      digest = 2

   class ResourceType(Enum):
      uri = 1
      stream = 2
      local = 3

   def __init__(self,service,cache=None,facets=None,graphs=None):
      self.logger = logging.getLogger(__name__)
      self.logger.setLevel(logging.DEBUG)
      self.logger.debug('Debug log for '+__name__)
      self.service = service
      self.cache = cache
      self.serviceAuth = self.authenticate('anonymous')
      self.resourceAuth = None
      self.resourceToken = None
      self.prefixes = { 'schema' : 'http://schema.org/'}
      self.graphs = []
      self.propertyTypes = {
         'url' : 'xsd:anyURI',
         'contentUrl' : 'xsd:anyURI',
         'isBasedOn' : 'xsd:anyURI',
         'isBasedOnUrl' : 'xsd:anyURI'
      }
      if graphs is not None:
         for graph in graphs:
            self.graphs.append(graph)

      self.facets = {}
      for facet in orgSchema:
         self.facets[facet.id] = facet
      if facets is not None:
         for facet in facets:
            self.facets[facet.id] = facet

   def isURIProperty(self,property):
      typename = self.propertyTypes.get(property)
      return True if typename=='xsd:anyURI' else False

   def getProlog(self):
      glob = io.StringIO()
      for key in self.prefixes:
         glob.write('prefix {}: <{}>'.format(key,self.prefixes[key]))
      return glob.getvalue()

   def setProxyCache(self,proxy,base):
      self.cache = { 'proxy': proxy, 'base':base }

   def authenticate(self,user,password='',method=AuthMethod.basic):
      if (method is Pond.AuthMethod.basic):
         self.auth = requests.auth.HTTPBasicAuth(user, password)
      elif (method is Pond.AuthMethod.digest):
         self.auth = requests.auth.HTTPDigestAuth(user, password)
      else:
         raise Exception('Unsupported authentication method')

   def defaultParameters(self):
      params = {}
      return params

   def mergeParameters(self,requestParameters):
      params = self.defaultParameters()
      for param in requestParameters.keys():
         params[param] = requestParameters[param]
      return params

   def getGraphs(self):
      named = ''
      for graph in self.graphs:
         named = named + 'from named <' + graph +'> '
      return named

   def facet(self,name,value = None):
      if type(value)==str:
         value = self.facets[value]
      return value if value is not None else self.facets[name]

   def currentEntity(self,entity = None,order = None,title = None, summary = None, criteria = None, name = None, genre = None):

      entityType = self.facet('type',entity)
      orderFacet = self.facet('order',order)
      titleFacet = self.facet('title',title)
      summaryFacet = self.facet('summary',summary)
      nameFacet = self.facet('name',name)
      genreFacet = self.facet('genre',genre)

      expr = '?s rdf:type {0}; {1} ?ordering; {2} ?title; {4} ?name; {5} ?genre; . optional {{ ?s {3} ?summary . }}'.format(entityType,orderFacet,titleFacet,summaryFacet,nameFacet,genreFacet)
      if criteria is not None:
         expr = expr + criteriaConditions(self,criteria)

      q = SPARQL() \
            .start(self.prefixes) \
            .select(['s','title','summary','ordering','name','genre']) \
            .fromGraphs(self.graphs) \
            .where(expr) \
            .orderBy('desc(?ordering)')
      if self.logger.isEnabledFor(logging.DEBUG):
         self.logger.debug(str(q))
      params = self.mergeParameters({'limit':1,'query':str(q)})

      req = requests.get(self.service,params=params,headers={'accept':'application/json'},auth=self.serviceAuth)

      if (req.status_code>=200 or req.status_code<300):
         #print(req.text)
         data = json.loads(req.text)
         if len(data['values']) == 0:
            return None
         subject = strip(data['values'][0][0])
         title = literal(data['values'][0][1])
         summary = literal(data['values'][0][2]) if data['values'][0][2] is not None else None
         dateTime = strip(data['values'][0][3])
         nameValue = strip(data['values'][0][4])
         genreValue = strip(data['values'][0][5])
         date,time = dateTime.split('T') # Need to check that is xsd:dateTime
         return (subject,title,summary,date,time,nameValue,genreValue)
      else:
         raise IOError('Cannot post data to uri <{}>, status={}'.format(self.service,req.status_code))

   def getEntityByName(self,value,name=None,title=None,order=None,summary=None,criteria=None,genre=None):
      nameFacet = self.facet('name',name)
      orderFacet = self.facet('order',order)
      titleFacet = self.facet('title',title)
      summaryFacet = self.facet('summary',summary)
      genreFacet = self.facet('genre',genre)

      expr = '?s {0} "{1}"; {2} ?ordering; {3} ?title; {5} ?genre; . optional {{ ?s {4} ?summary . }}'.format(nameFacet,value,orderFacet,titleFacet,summaryFacet,genreFacet)
      if criteria is not None:
         expr = expr + criteriaConditions(self,criteria)

      q = SPARQL() \
            .start(self.prefixes) \
            .select(['s','title','summary','ordering','genre']) \
            .fromGraphs(self.graphs) \
            .where(expr) \
            .orderBy('desc(?ordering)')
      if self.logger.isEnabledFor(logging.DEBUG):
         self.logger.debug(str(q))
      params = self.mergeParameters({'limit':1,'query':str(q)})

      req = requests.get(self.service,params=params,headers={'accept':'application/json'},auth=self.serviceAuth)

      if (req.status_code>=200 or req.status_code<300):
         #print(req.text)
         data = json.loads(req.text)
         if len(data['values']) == 0:
            return None
         subject = strip(data['values'][0][0])
         title = literal(data['values'][0][1])
         summary = literal(data['values'][0][2]) if data['values'][0][2] is not None else None
         dateTime = strip(data['values'][0][3])
         date,time = dateTime.split('T') # Need to check that is xsd:dateTime
         genreValue = strip(data['values'][0][4])
         return (subject,title,summary,date,time,value,genreValue)
      else:
         raise IOError('Cannot query uri <{}>, status={}'.format(self.service,req.status_code))

   def proxyURL(self,resource):
      uri = resource
      if not self.cache==None and 'proxy' in self.cache and 'base' in self.cache:
         base = self.cache['base']
         relative = resource[resource.index(base) + len(base):]
         uri = self.cache['proxy'] + relative
      if not self.cache==None and 'dir' in self.cache:
         relative = resource
         if 'base' in self.cache:
            base = self.cache['base']
            relative = resource[resource.index(base) + len(base):]
         uri = 'file://' + os.path.join(os.path.abspath(self.cache['dir']),relative)
      return uri

   def getResourceText(self,resource,authorization=None):

      uri = self.proxyURL(resource)

      self.logger.debug(uri)

      if self.cache==None or uri[0:5]=='http:' or uri[0:6]=='https:':
         headers = {}
         if self.resourceToken is not None:
            headers['Authorization'] = 'Bearer ' + self.resourceToken
         req = requests.get(uri,auth=self.resourceAuth,headers=headers)
         if (req.status_code == 200):
            contentType = req.headers.get('Content-Type')
            encoding = 'UTF-8'
            if contentType is not None:
               param = contentType.find('charset=')
               if param>0:
                  encoding = contentType[param+8:]
            text = req.content.decode(encoding)
            return text
         else:
            raise IOError('Cannot get <{}>, status={}'.format(uri,req.status_code))
      elif uri[0:7]=='file://':
         entryFile = open(uri[7:], mode='r', encoding='utf-8')
         entry = entryFile.read()
         entryFile.close()
         return entry
      else:
         entryFile = open(uri, mode='r', encoding='utf-8')
         entry = entryFile.read()
         entryFile.close()
         return entry

   def getResource(self,resource):

      uri = self.proxyURL(resource)

      if self.cache!=None and 'redirect' in self.cache and self.cache['redirect']:
         return (Pond.ResourceType.uri,uri,303)
      if self.cache==None or uri[0:5]=='http:' or uri[0:6]=='https:':
         headers = {}
         if self.resourceToken is not None:
            headers['Authorization'] = 'Bearer ' + self.resourceToken
         req = requests.get(uri, stream = True,auth=self.resourceAuth,headers=headers)
         return (Pond.ResourceType.stream,req.iter_content(20*1024),req.headers['content-type'],req.headers.get('content-length'))
      elif uri[0:7]=='file://':
         location = uri[7:]
         return (Pond.ResourceType.local,location,None,os.path.getsize(location))
      else:
         return (Pond.ResourceType.local,uri,None,os.path.getsize(uri))

   def relatedEntityByOrder(self,value,previous = True,limit = 1,singleton = False,entity = None,order = None,title = None, summary = None, criteria = None, name=None, genre=None):
      entityType = self.facet('type',entity)
      orderFacet = self.facet('order',order)
      titleFacet = self.facet('title',title)
      summaryFacet = self.facet('summary',summary)
      nameFacet = self.facet('name',name)
      genreFacet = self.facet('genre',genre)
      expr = '?s rdf:type {0}; {1} ?ordering; {2} ?title; {4} ?name; {5} ?genre; . optional {{ ?s {3} ?summary . }}'.format(entityType,orderFacet,titleFacet,summaryFacet,nameFacet,genreFacet) \
             + ' FILTER( ?ordering ' + ('<' if previous else '>') + ' ' + orderFacet.toLiteral(value) + ' )'
      if criteria is not None:
         expr = expr + criteriaConditions(self,criteria)
      q = SPARQL() \
            .start(self.prefixes) \
            .select(['s','title','summary','ordering','name','genre']) \
            .fromGraphs(self.graphs) \
            .where(expr ) \
            .orderBy('desc(?ordering)' if previous else 'asc(?ordering)')
      if self.logger.isEnabledFor(logging.DEBUG):
         self.logger.debug(str(q))
      params = self.mergeParameters({'limit':limit,'query':str(q)})

      req = requests.get(self.service,params=params,headers={'accept':'application/json'},auth=self.serviceAuth)

      if (req.status_code>=200 or req.status_code<300):
         data = json.loads(req.text)
         values = data['values']
         if len(values)==0:
            return None
         elif len(values)==1 and not singleton:
            row = values[0]
            subjectValue = strip(row[0])
            titleValue = literal(row[1])
            summaryValue = literal(row[2]) if row[2] is not None else None
            date,time = strip(row[3]).split('T')
            nameValue = strip(row[4])
            genreValue = strip(row[5])
            return (subjectValue,titleValue,summaryValue,date,time,nameValue,genreValue)
         else:
            result = []
            for row in values:
               subjectValue = strip(row[0])
               titleValue = literal(row[1])
               summaryValue = literal(row[2]) if row[2] is not None else None
               date,time = strip(row[3]).split('T')
               nameValue = strip(row[4])
               genreValue = strip(row[5])
               result.append((subjectValue,titleValue,summaryValue,date,time,nameValue,genreValue))
            return result
      else:
         raise IOError('Cannot post data to uri <{}>, status={}'.format(self.service,req.status_code))

   def entityByOrder(self,value,entity = None,order = None,title = None,summary = None,resource = None, criteria = None, name=None, genre=None):
      return self.entityByValue(value,self.facet('order',order),entity=entity,order=order,title=title,summary=summary,resource=resource,criteria=criteria,name=name,genre=genre)

   def entityByValue(self,value,facet,entity = None,order = None,title = None,summary = None,resource = None, criteria = None, name=None, genre=None):
      entityType = self.facet('type',entity)
      orderFacet = self.facet('order',order)
      titleFacet = self.facet('title',title)
      summaryFacet = self.facet('summary',summary)
      nameFacet = self.facet('name',name)
      genreFacet = self.facet('genre',genre)
      expr = '?s {0} ?ordering; {1} ?title; {3} {4}; {5} ?name; {6} ?genre . optional {{ ?s {2} ?summary . }}'.format(orderFacet,titleFacet,summaryFacet,facet,facet.toLiteral(value),nameFacet,genreFacet)
      if criteria is not None:
         expr = expr + criteriaConditions(self,criteria)
      q = SPARQL() \
            .start(self.prefixes) \
            .select(['s','title','summary','ordering','name','genre']) \
            .fromGraphs(self.graphs) \
            .where(expr)
      if self.logger.isEnabledFor(logging.DEBUG):
         self.logger.debug(str(q))
      params = self.mergeParameters({'limit':1,'query':str(q)})

      req = requests.get(self.service,params=params,headers={'accept':'application/json'},auth=self.serviceAuth)

      if (req.status_code>=200 or req.status_code<300):
         data = json.loads(req.text)
         values = data['values']
         if len(values)==0:
            return None
         subject = strip(values[0][0])
         title = literal(values[0][1])
         summary = literal(values[0][2]) if values[0][2] is not None else None
         dateTime = strip(values[0][3])
         date,time = dateTime.split('T') # Need to check that is xsd:dateTime
         nameValue = strip(values[0][4])
         genreValue = strip(values[0][5])
         return (subject,title,summary,date,time,nameValue,genreValue)
      else:
         raise IOError('Cannot post data to uri <{}>, status={}'.format(self.service,req.status_code))

   def getCategoryCount(self,entity = None,category = None):
      entityType = self.facet('type',entity)
      categoryFacet = self.facet('category',category)
      expr = '?s rdf:type {0}; {1} ?category'.format(entityType,categoryFacet)
      q = SPARQL() \
            .start(self.prefixes) \
            .select(['category','(count(?category) as ?count)']) \
            .fromGraphs(self.graphs) \
            .where(expr) \
            .groupBy('?category') \
            .orderBy('desc(?count)')
      if self.logger.isEnabledFor(logging.DEBUG):
         self.logger.debug(str(q))
      params = self.mergeParameters({'limit':100,'query':str(q)})

      req = requests.get(self.service,params=params,headers={'accept':'application/json'},auth=self.serviceAuth)

      if (req.status_code>=200 or req.status_code<300):
         data = json.loads(req.text)
         keywords = {}
         for keyword in data['values']:
            value,type = keyword[1].split('^^')
            keywords[strip(keyword[0])] = int(strip(value))
         return keywords
      else:
         raise IOError('Cannot post data to uri <{}>, status={}'.format(self.service,req.status_code))

   def getEntityCategories(self,subject,entity = None,category = None):
      entityType = self.facet('type',entity)
      categoryFacet = self.facet('category',category)
      expr = '<{0}> rdf:type {1}; {2} ?category'.format(subject,entityType,categoryFacet)
      q = SPARQL() \
            .start(self.prefixes) \
            .select(['category']) \
            .fromGraphs(self.graphs) \
            .where(expr)
      if self.logger.isEnabledFor(logging.DEBUG):
         self.logger.debug(str(q))
      params = self.mergeParameters({'limit':100,'query':str(q)})

      req = requests.get(self.service,params=params,headers={'accept':'application/json'},auth=self.serviceAuth)

      if (req.status_code>=200 or req.status_code<300):
         data = json.loads(req.text)
         keywords = []
         for keyword in data['values']:
            keywords.append(strip(keyword[0]))
         return keywords
      else:
         raise IOError('Cannot post data to uri <{}>, status={}'.format(self.service,req.status_code))

   def getEntitiesByCategory(self,value,limit = 100,entity = None,category = None,title = None,summary = None,order = None,resource = None, name=None, genre=None):
      entityType = self.facet('type',entity)
      categoryFacet = self.facet('category',category)
      titleFacet = self.facet('title',title)
      summaryFacet = self.facet('summary',summary)
      orderFacet = self.facet('order',order)
      nameFacet = self.facet('name',name)
      genreFacet = self.facet('genre',genre)
      expr = '?s rdf:type {0}; {1} ?title; {3} ?date; {4} {5}; {6} ?name; {7} ?genre; . optional {{ ?s {2} ?summary; }}'.format(entityType,titleFacet,summaryFacet,orderFacet,categoryFacet,categoryFacet.toLiteral(value),nameFacet,genreFacet)
      q = SPARQL() \
            .start(self.prefixes) \
            .select(['s','title','summary','date','name','genre']) \
            .fromGraphs(self.graphs) \
            .where(expr)
      if self.logger.isEnabledFor(logging.DEBUG):
         self.logger.debug(str(q))
      params = self.mergeParameters({'limit':limit,'query':str(q)})

      req = requests.get(self.service,params=params,headers={'accept':'application/json'},auth=self.serviceAuth)

      if (req.status_code>=200 or req.status_code<300):
         data = json.loads(req.text)
         result = []
         for row in data['values']:
            subjectValue = strip(row[0])
            titleValue = literal(row[1])
            summaryValue = literal(row[2]) if row[2] is not None else None
            date,time = strip(row[3]).split('T')
            nameValue = strip(row[4])
            genreValue = strip(row[5])
            result.append((subjectValue,titleValue,summaryValue,date,time,nameValue,genreValue))
         return result
      else:
         raise IOError('Cannot post data to uri <{}>, status={}'.format(self.service,req.status_code))

   def getEntityResource(self,subject,singleton = False,resource=None,contentUrl=None,fileFormat=None,criteria=None):
      resourceFacet = self.facet('resource',resource)
      if resourceFacet.type=='schema:MediaObject':
         contentUrlFacet = self.facet('contentUrl',contentUrl)
         fileFormatFacet = self.facet('fileFormat',fileFormat)
         expr = '<{0}> {1} ?r . ?r rdf:type ?type; {2} ?url; {3} ?format .'.format(subject,resourceFacet,contentUrlFacet,fileFormatFacet)
         if criteria is not None:
            expr = expr + criteriaConditions(self,criteria,subjectVar='r')
         q = SPARQL() \
               .start(self.prefixes) \
               .select(['url','format','type']) \
               .fromGraphs(self.graphs) \
               .where(expr)
         if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(str(q))
         params = self.mergeParameters({'query':str(q)})
         req = requests.get(self.service,params=params,headers={'accept':'application/json'},auth=self.serviceAuth)

         if (req.status_code>=200 or req.status_code<300):
            data = json.loads(req.text)
            values = data['values']
            if len(values)==0:
               return None
            elif len(values)==1 and not singleton:
               row = values[0]
               return (strip(row[0]),strip(row[1]),strip(row[2]))
            else:
               result = []
               for row in values:
                  result.append((strip(row[0]),strip(row[1]),strip(row[2])))
               return result
         else:
            raise IOError('Cannot get data from uri <{}>, status={}'.format(self.service,req.status_code))
      elif resourceFacet.type=='xsd:anyURI':
         expr = '<{0}> {1} ?url.'.format(subject,resourceFacet)
         if criteria is not None:
            expr = expr + criteriaConditions(self,criteria,subjectVar='r')
         q = SPARQL() \
               .start(self.prefixes) \
               .select(['url']) \
               .fromGraphs(self.graphs) \
               .where(expr)
         if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(str(q))
         params = self.mergeParameters({'query':str(q)})
         req = requests.get(self.service,params=params,headers={'accept':'application/json'},auth=self.serviceAuth)
         if (req.status_code>=200 or req.status_code<300):
            data = json.loads(req.text)
            values = data['values']
            if len(values)==0:
               return None
            elif len(values)==1 and not singleton:
               row = values[0]
               return (strip(row[0]),None,None)
            else:
               result = []
               for row in values:
                  result.append((strip(row[0]),None,None))
               return result
         else:
            raise IOError('Cannot get data from uri <{}>, status={}'.format(self.service,req.status_code))
      else:
         raise IOError('Cannot determine resource facet query, type {}'.format(resourceFacet.type))



from flask_wtf import FlaskForm
from wtforms import StringField, BooleanField
from wtforms.validators import DataRequired

class LoginForm(FlaskForm):
    username = StringField('username', validators=[DataRequired()])
    password = StringField('password', validators=[DataRequired()])
    redirectonlogin = StringField('redirectonlogin', default='true',validators=[])



from flask import Blueprint
from flask import send_from_directory, render_template, abort, current_app, Response, stream_with_context
import requests
import os, io

docs = Blueprint('duckpond_blog_docs',__name__)

@docs.route('/docs/<path:path>')
def send_doc(path):
   siteURL = current_app.config.get('SITE_URL')

   templateOptions = current_app.config.get('TEMPLATE_OPTIONS')
   if templateOptions is None:
      templateOptions = {
         'title' : 'My Journal'
      }

   location = current_app.config.get('DOCS')
   if location is None:
      abort(404)

   if location[0:4]=='http':
      url = location + path
      req = requests.get(url, stream = True,headers={'Connection' : 'close'})
      if req.headers['Content-Type'][0:9]=='text/html':
         return render_template('content.html', siteURL=siteURL if siteURL is not None else request.url_root[0:-1], options=templateOptions, html=req.text, entry=None)
      else:
         return Response(stream_with_context(req.iter_content()), headers = dict(req.headers))

   else:

      dir = os.path.abspath(location)
      if path.endswith('.html'):

         glob = io.StringIO()
         try:
            with open(os.path.join(dir,path), mode='r', encoding='utf-8') as doc:
               peeked = doc.readline()
               if peeked.startswith('<!DOCTYPE'):
                  return send_from_directory(dir, path)
               glob.write(peeked)
               for line in doc:
                  glob.write(line)

               return render_template('content.html', siteURL=siteURL if siteURL is not None else request.url_root[0:-1], options=templateOptions, html=glob.getvalue(), entry=None)
         except FileNotFoundError:
            abort(404)

      return send_from_directory(dir, path)
